@@ -106,7 +106,9 @@ def construct_MG(faces, n_vertices, gamma):
 
 def NRICP(floating, fixed, fl_landmarks=None, fi_landmarks=None, verbose=False, mask=None):
     
-    """ Perform non-rigid ICP to register floating mesh to fixed 
+    """ Perform optimal step non-rigid ICP to register floating mesh to fixed 
+        (Amberg & Romdhani)
+        https://www.researchgate.net/publication/200172513_Optimal_Step_Nonrigid_ICP_Algorithms_for_Surface_Registration
     
         (params):
             floating (Mesh): The mesh to be aligned
@@ -129,28 +131,6 @@ def NRICP(floating, fixed, fl_landmarks=None, fi_landmarks=None, verbose=False,
     fl_verts = floating.vertices
     fi_verts = fixed.vertices
     
-    """
-    # Initially scale into the unit cube
-    fl_max, fl_min = np.max(fl_verts, axis=0), np.min(fl_verts, axis=0)
-    fi_max, fi_min = np.max(fi_verts, axis=0), np.min(fi_verts, axis=0)
-    
-    # [0,1] scaling
-    fl_verts = (fl_verts - fl_min) / (fl_max - fl_min)
-    fi_verts = (fi_verts - fi_min) / (fi_max - fi_min)
-    
-    # [-1,1] scaling
-    fl_verts = (2 * fl_verts) - 1
-    fi_verts = (2 * fi_verts) - 1
-    
-    print(np.max(fl_verts, axis=0), np.min(fl_verts,axis=0), (np.max(fl_verts, axis=0)+np.min(fl_verts, axis=0))/2)
-    print(np.max(fi_verts, axis=0), np.min(fi_verts,axis=0), (np.max(fi_verts, axis=0)+np.min(fi_verts, axis=0))/2)
-    """
-    
-    if verbose:
-        print('Scaled data to: ')
-        print(np.min(fl_verts), np.max(fl_verts))
-        print(np.min(fi_verts), np.max(fi_verts))
-    
     n_vertices = fl_verts.shape[0]
     n_vertices_target = fi_verts.shape[0]
     
@@ -221,7 +201,7 @@ def NRICP(floating, fixed, fl_landmarks=None, fi_landmarks=None, verbose=False,
         
         UL = fi_landmark_points
 
-    # -------------------- Loop over stiffness ------------------------ #
+    # ----------------------- Loop over stiffness --------------------------- #
     
     graph = MeshGraph(floating)
     
@@ -240,8 +220,10 @@ def NRICP(floating, fixed, fl_landmarks=None, fi_landmarks=None, verbose=False,
         
         # Keep track of changes in the transform
         change = np.inf
-        n_iter = 0 # Early stoconda install -c conda-forge scikit-sparsep if too many iterations
+        n_iter = 0 # Early stop if too many iterations
         
+        
+        # ------------------ Loop until convergence ------------------------- #
         while change > CHANGE_THRESHOLD:
             
             weights = np.ones((n_vertices,1))
@@ -249,20 +231,19 @@ def NRICP(floating, fixed, fl_landmarks=None, fi_landmarks=None, verbose=False,
             transformed = D @ transform
             
             dist, idx = nn.kneighbors(transformed[:,0:3])
-            idx = idx.squeeze()   # Not sure why but https://github.com/saikiran321/nonrigid_icp/blob/master/nricp.py does it
+            idx = idx.squeeze()   # https://github.com/saikiran321/nonrigid_icp/blob/master/nricp.py does this and it helps
             correspondance = fi_verts[idx, 0:3]
             
             # Discard any poor matches
             weights[np.where(dist>DISCARD)] = 0 
             
+            # If using normals for weighting
             if USE_NORMALS and alpha < 3:
-                
+                # Only do it for low stiffness to overcome poor initial registration
                 dists = np.zeros(fl_verts.shape[0])
                 norms = get_normals(transformed , floating.faces)
                 dist_k, idx_k = knn.kneighbors(transformed)
-                
                 for i in range(fl_verts.shape[0]):
-                    
                     n = norms[i]
                     k_corr = fi_verts[idx_k[i], 0:3] # This gives the KNN of the point
                     diff = fl_verts[i, 0:3] - k_corr # This gets the direction of each KNN relative to the point
@@ -283,9 +264,10 @@ def NRICP(floating, fixed, fl_landmarks=None, fi_landmarks=None, verbose=False,
                 for idx in mask:
                     weights[idx] = 0
                     
-            
+            # Weighted match matrix
             U = weights * correspondance
             
+            # A,B,U and D Directly from the paper
             A = sparse.vstack([alpha * MG, D.multiply(weights)])
             A = sparse.csr_matrix(A)
             
@@ -294,21 +276,22 @@ def NRICP(floating, fixed, fl_landmarks=None, fi_landmarks=None, verbose=False,
             B = sparse.csr_matrix(B)
             
             if USE_LANDMARKS:
-                
+                # Landmark terms + weighting
                 DL = sparse.lil_matrix((UL.shape[0], 4*n_vertices))
                 for count, l_idx in enumerate(fl_landmark_indices):
                     DL[count, 4*l_idx: 4*l_idx + 4] = fl_verts[l_idx,:] * beta                    
             
                 A = sparse.csr_matrix(sparse.vstack([A, DL]))
                 B = sparse.csr_matrix(sparse.vstack([B, beta * UL]))
-                
+            
+            # This seems to be the only way of solving without lots of RAM    
             factor = cholesky_AAt(A.T)
             X = factor(A.T.dot(B)).toarray()
             
             change = np.linalg.norm(X-transform)
             transform = X
             n_iter += 1
-            if n_iter > MAX_ITERS:
+            if n_iter > MAX_ITERS: # Early stop
                 break
             
             if verbose:
@@ -322,13 +305,14 @@ def NRICP(floating, fixed, fl_landmarks=None, fi_landmarks=None, verbose=False,
                 
                 else:
                     print('Loss: ', get_d_loss(D@transform, correspondance) + alpha * get_s_loss(graph, transform, 1.0))
-            
+    
+    # Copy face and UV information from the ols mesh        
     out_faces = floating.faces
     out_uvs = floating.UVs
     out_normals = floating.normals
     
     out_vertices = D @ transform
-
+    
     centroid = np.mean(out_vertices, axis=0)
     for i in range(n_vertices):
         
@@ -342,24 +326,12 @@ def NRICP(floating, fixed, fl_landmarks=None, fi_landmarks=None, verbose=False,
         # If we are told to, try to flip some incorrect normals
         if FIX_NORMALS:
             
-            # Get the line to center
-            d_orig = out_vertices[i,:] - centroid
-            d_orig /= np.linalg.norm(d_orig)
-            # If this line is more than 90' off the normal, flip the normal
-            cos_theta = d_orig.dot(n_)
-            if cos_theta < 0:
-                n_ *= -1
-                print('Flipped normal ', i)
-            
+            # TODO: Fix backface issues
+            # We'll look at the nearest neighbours and if more
+            # than a certain ratio are opposite in direction we flip
+            pass
+        
         out_normals[i, :] = n_
-    
-    """        
-    # Rescale to the original size
-    out_vertices = (out_vertices + 1)/2
-    out_vertices = (out_vertices * (fl_max - fl_min)) + fl_min   
-    
-    print(np.max(out_vertices, axis=0), np.min(out_vertices,axis=0), (np.max(out_vertices, axis=0)+np.min(out_vertices, axis=0))/2)
-    """
     return Mesh(out_vertices, out_uvs, out_normals, out_faces), transform
             
             
