@@ -1,4 +1,4 @@
-from Mesh.Mesh import Mesh
+from PyFaceGen.Mesh.Mesh import Mesh
 
 class MeshGraph:
     
@@ -35,4 +35,4 @@ class MeshGraph:
             print(line)
 
     
-    
\ No newline at end of file
+    
